#include <stdio.h>
#include <fcntl.h>
#include <dahdi/user.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <pcap.h>
#include <sys/types.h>
#include <arpa/inet.h>

#ifndef ETH_P_LAPD
short ETH_P_LAPD;
#endif

#define BLOCK_SIZE 512



struct mtp2_phdr {
        u_int8_t  sent;
        u_int8_t  annex_a_used;
        u_int16_t link_number;
};

struct lapd_sll_hdr {
    u_int16_t sll_pkttype;    /* packet type */
    u_int16_t sll_hatype;
    u_int16_t sll_halen;
    u_int8_t sll_addr[8];
    u_int16_t sll_protocol;   /* protocol, should be ETH_P_LAPD */
};

struct chan_fds {
	int rfd;
	int tfd;
	int chan_id;
	char tx_buf[BLOCK_SIZE * 4];
	int tx_len;
	char rx_buf[BLOCK_SIZE * 4];
	int rx_len;
};

int make_mirror(long type, int chan)
{
	int res = 0;
	int fd = 0;	
	int x = BLOCK_SIZE;
	struct dahdi_bufferinfo bi;
	fd = open("/dev/dahdi/pseudo", O_RDONLY);

	memset(&bi, 0, sizeof(bi));
        bi.txbufpolicy = DAHDI_POLICY_IMMEDIATE;
        bi.rxbufpolicy = DAHDI_POLICY_IMMEDIATE;
        bi.numbufs = 32;
        bi.bufsize = BLOCK_SIZE;

	ioctl(fd, DAHDI_SET_BUFINFO, &bi);

	res = ioctl(fd, type, &chan);
	if(res)
	{
		printf("error setting channel err=%d!\n", res);
		return -1;
	}
	
	return fd;
}

void log_packet(struct chan_fds * fd, char is_read, pcap_dumper_t * dump)
{
	char buf[BLOCK_SIZE * 4];
	int res = 0;

	struct pcap_pkthdr hdr;
	struct lapd_sll_hdr * lapd = (struct lapd_sll_hdr *)buf;
	memset(buf, 0, sizeof(buf));
	if(is_read)
	{
		res = read(fd->rfd, buf+sizeof(struct lapd_sll_hdr), sizeof(buf)-sizeof(struct lapd_sll_hdr));

		if(fd->rx_len > 0 && res == fd->rx_len && !memcmp(fd->rx_buf, buf+sizeof(struct lapd_sll_hdr), res) )
		{
			//skipping dup
//			return;
		}

		memcpy(fd->rx_buf,  buf+sizeof(struct lapd_sll_hdr), res);
		fd->rx_len = res;
	}
	else
	{
		res = read(fd->tfd, buf+sizeof(struct lapd_sll_hdr), sizeof(buf)-sizeof(struct lapd_sll_hdr));
		if(fd->tx_len > 0 && res == fd->tx_len && !memcmp(fd->tx_buf, buf+sizeof(struct lapd_sll_hdr), res) )
		{
			//skipping dup
//			return;
		}

		memcpy(fd->tx_buf,  buf+sizeof(struct lapd_sll_hdr), res);
		fd->tx_len = res;
	}

/*
	if(!buf[sizeof(struct lapd_sll_hdr)+2])
	{	
		// skipping FISSU

		return;
	}
*/


	gettimeofday(&hdr.ts, NULL);

	

	
	if(res > 0)
	{
		hdr.caplen = res+sizeof(struct lapd_sll_hdr)-2;
		hdr.len = res+sizeof(struct lapd_sll_hdr)-2;
		
		int i = 0;
		lapd->sll_pkttype = 3;
		lapd->sll_hatype = 0;
		lapd->sll_halen = res;
//                lapd->sll_addr = ???
                lapd->sll_protocol = ETH_P_LAPD;

		pcap_dump((u_char*)dump, &hdr, buf);
		pcap_dump_flush(dump);
	}
}

int main(int argc, char *argv[])
{
	struct chan_fds chans[10];
	int num_chans = argc - 2;
	int max_fd = 0;

	int i;	

ETH_P_LAPD = htons(0x0030);

	for(i = 0; i < num_chans; i++)
	{
		int chan = atoi(argv[i+1]);

		chans[i].tfd = make_mirror(DAHDI_TXMIRROR, chan);
		chans[i].rfd = make_mirror(DAHDI_RXMIRROR, chan);
		chans[i].chan_id = chan;
		chans[i].tx_len = 0;
		chans[i].rx_len = 0;
		if(chans[i].tfd > max_fd)
		{
			max_fd = chans[i].tfd;
		}
		if(chans[i].rfd > max_fd)
		{
			max_fd = chans[i].rfd;
		}
	}
	max_fd++;
	

	pcap_t * pcap = pcap_open_dead(DLT_LINUX_LAPD, BLOCK_SIZE*4);
	pcap_dumper_t * dump = pcap_dump_open(pcap, argv[argc - 1]);
	while(1)
	{
		fd_set rd_set;
		FD_ZERO(&rd_set);
		for(i = 0; i < num_chans; i++)
		{
			FD_SET(chans[i].tfd, &rd_set);
			FD_SET(chans[i].rfd, &rd_set);
		}

		select(max_fd, &rd_set, NULL, NULL, NULL);

		for(i = 0; i < num_chans; i++)
		{
			if(FD_ISSET(chans[i].rfd, &rd_set))
			{
				log_packet(&chans[i], 1, dump);
			}
			if(FD_ISSET(chans[i].tfd, &rd_set))
			{
				log_packet(&chans[i], 0, dump);
			}
		}
	}

	return 0;
}
